@@ -383,9 +383,10 @@ int vm_run(META *m, VM *vm)
 				msg(F_SEND, L_DEBUG, "OP_DICTENC: tree on ground space %s:\n", R.i->spc->name);
 				meta_printavlist(m, avlst, 0);
 			}
-			/* Now encode it */
+			/* Now encode it and free the tree */
 			NEWSTR(s, C_MAX_PKTSIZE); 
 			R.l = meta_encode(R.i->spc, s, C_MAX_PKTSIZE, avlst, 0);
+			meta_freeavtree(avlst);
 			if (R.l < 0) { R.l = 0; free(s); break; }
 			R.p = s; R.i = 0; R.flags |= AV_FREE_P;
 			if (msg_thresh[F_SEND] >= L_DEBUG) hexdump(s, R.l);
@@ -485,8 +486,10 @@ int vm_run(META *m, VM *vm)
 			else { meta_freeavdata(&L); L.l = 0; }
 			break;
 		  case OP_FO:
+			/* Move R to L so that R can be freed */
 		  	n = L.l; meta_freeavdata(&L); 
 			L = R; R.flags &= ~AV_FREE_P;
+			/* We only adjust the length here; nothing else */
 			if ((META_SORD)n < 0) n = L.l + n;
 			if (n >= 0 && n < L.l) L.l = n;
 			break;
@@ -495,6 +498,9 @@ int vm_run(META *m, VM *vm)
 			L = R; R.flags &= ~AV_FREE_P;
 			if ((META_SORD)n < 0) n = L.l + n;
 			if (n >= 0 && n < L.l) {
+				/* We always copy; we never advance the
+				 * pointer, for now. And we only copy in
+				 * place if the stack item owns the data. */
 				if (L.flags & AV_FREE_P) s = L.p;
 				else { NEWSTR(s, n); L.flags |= AV_FREE_P; }
 				memcpy(s, L.p + (L.l - n), n);
@@ -511,9 +517,15 @@ int vm_run(META *m, VM *vm)
 			break;
 		  case OP_CONCAT:
 		  	if (!R.l) break;
-			if (!L.l) { L = R; break; }
+#if 0	/* FIXME: only commented out to allow copy-by-noop-concat hack */
+			if (!L.l) {
+				meta_freeavdata(&L);
+				L = R; R.flags &= ~AV_FREE_P; 
+				break; 
+			}
+#endif
 			NEWSTR(s, L.l + R.l);
-			memcpy(s, L.p, L.l);
+			if (L.l) memcpy(s, L.p, L.l);
 			memcpy(s + L.l, R.p, R.l);
 			CHGSTR(&L, s, L.l + R.l);
 			break;
@@ -851,29 +863,40 @@ int vm_run(META *m, VM *vm)
 				L.flags &= ~(AV_UNRESOLVED|AV_FREE_P);
 			}
 
-			/* When changing the underlying item's data, we
-		   	   copy the *data* from the right term, unless the
-			   right term is anonymous and non-temporary (this
-			   means its data is immediate). 
-			   We don't re-use non-temporary pointers from
-			   stack items that do have underlying items, as they 
-			   could point to their underlying items' data, which 
-			   is freed when the item itself is deleted! */
-			if (L.sub != R.sub) meta_freeavdata(L.sub);
-			L.sub->l = R.l;
-			if (!R.i && !(R.flags & AV_FREE_P)) {
+			/* Deal with the existing or newly created list item.
+			 *
+			 * We copy the data from the right hand stack item into
+			 * list item owned memory in all cases except the 
+			 * following:
+			 * a. the data is owned by the code (stack item is 
+			 *    both anonymous and does not own data), or
+			 * b. the data is owned by the request (stack item
+			 *    pointer is same as list item pointer and list
+			 *    item does not own data), or
+			 * c. the data is owned by the list item, and the
+			 *    source and destination list item are the same.
+			 *    This can currently occur only after a firstof
+			 *    or lastof on a non-anonymous stack item whose
+			 *    data is owned by its underlying list item.
+			 */
+
+			if ((!R.i && !(R.flags & AV_FREE_P)) ||	  /* code */
+			    (R.i && R.sub && R.p == R.sub->p &&   /* req/same */
+			    (!(R.sub->flags & AV_FREE_P) || R.sub == L.sub))) {
+				if (L.sub != R.sub) meta_freeavdata(L.sub);
 				L.sub->p = R.p;
+				L.sub->l = R.l;
 			}
 			else if (R.l && R.p) {
-				NEWSTR(L.sub->p, R.l);
-				memcpy(L.sub->p, R.p, R.l);
-				L.sub->flags |= AV_FREE_P;
+				NEWSTR(s, R.l);
+				memcpy(s, R.p, R.l);
+				CHGSTR(L.sub, s, R.l);
 			}
-			else L.sub->p = 0;
+			else L.sub->p = 0, L.sub->l = R.l;
 
-			/* Now change the left operand as if we would have
+			/* Now change the left stack item as if we would have
 			   re-done the fixup at this point. Free any possible
-			   temporary data held by the stack item first though */
+			   temporary data owned by the stack item first. */
 			meta_freeavdata(&L);
 			L.l = L.sub->l;
 			L.p = L.sub->p;
