@@ -67,7 +67,14 @@ char eaptls_id[] = "EAPTLS - Copyright (C) 2005 Emile van Bergen.";
 
 /* Dictionary items not in common/constants.h */
 
-#define C_DI_FRAMED_MTU		12
+#define C_DI_FRAMED_MTU		12	/* In C_DS_RAD_ATTR */
+
+#define C_DI_TLS_PAYLOAD	51	/* In C_DS_INTERNAL */
+#define C_DI_TLS_RESULT		52	
+#define C_DI_TLS_ACTION		53
+
+#define C_DV_TLS_FLAGS_MORE	0x40
+#define C_DV_TLS_FLAGS_LENGTH	0x80
 
 
 /*
@@ -84,8 +91,14 @@ typedef struct conv {
     int inuse;
     conv_state_t state;
     time_t exp;
+
     gnutls_session_t session;
     BUF_T fromtls, totls;	    /* transport buffers (below TLS) */
+    BUF_T totls_app;		    /* application buffer (above TLS);
+				       we don't buffer application dat
+				       from TLS, as we send that to OR
+				       straight away */ 
+    int result;
 } CONV_T;
 
 
@@ -94,9 +107,9 @@ typedef struct conv {
  */
 
 
-char *buf[TLS_BUFFER_SIZE];	    /* scratch buffer */
+char buf[TLS_BUFFER_SIZE];	    /* scratch buffer */
 
-META_ORD rxspace, txspace;
+META_ORD rx_space, tx_space;
 int debug;
 
 
@@ -113,8 +126,8 @@ ssize_t tlspull(gnutls_transport_ptr_t td, void *data, size_t len)
     CONV_T *conv = (CONV_T *)td;
     if (!conv->inuse) { errno = EBADF; return -1; }
 
-    /* Another copy, for no reason outside Gnutls' design. And then Gnutls
-     * proceeds to make a few more of its own. Sigh. */
+    /* Another copy, and then Gnutls proceeds to make a few more of its own.
+     * Sigh. */
 
     if (len > buf_maxget(&conv->totls)) { errno = EAGAIN; return -1; }
     memcpy(data, conv->totls.r, len);
@@ -138,8 +151,8 @@ ssize_t tlspush(gnutls_transport_ptr_t td, const void *data, size_t len)
 
     if (len > buf_maxput(&conv->fromtls)) { errno = ENOSYS; return -1; }
 
-    /* Another copy, for no reason outside Gnutls' design. And then Gnutls
-     * proceeds to make a few more of its own. Sigh. */
+    /* Another copy, and then Gnutls proceeds to make a few more of its own.
+     * Sigh. */
 
     memcpy(conv->fromtls.w, data, len);
     buf_put(&conv->fromtls, len);
@@ -177,13 +190,13 @@ void parseoptions(int argc, char **argv)
     /* Handle options */
     while((c = getopt(argc, argv, "dhvr:s:")) != -1) {
         switch(c) {
-          case 's': rxspace = txspace = atoi(optarg); 
-		    if (!txspace) {
+          case 's': rx_space = tx_space = atoi(optarg); 
+		    if (!tx_space) {
 			fprintf(stderr, "radldap: Invalid space number '%s'!\n",
 				optarg);
 			usage();
 		    }
-		    rxspace++;
+		    rx_space++;
 		    break;
           case 'd': debug++; break;
           case 'v': fprintf(stderr, "\nEAPTLS module " MODULE_VERSION ". "
@@ -229,58 +242,106 @@ static inline int addint(uint32_t **p, uint32_t atrnr, uint32_t val)
 /* Proceed with TLS conversation according to its state; may add
  * attributes to module response. */
 
-uint32_t tls_state_machine(uint32_t **rep, CONV_T *conv)
+typedef enum { RETURN, APP_RECV, IA_RECV, PRERROR, ERROR, DUMPSESS } action_t;
+
+uint32_t tls_state_machine(CONV_T *conv, uint32_t **rep, uint32_t *repend,
+			   int app_action)
 {
-    static 
+    action_t action;
     int ret, gr;
+    uint32_t *o;
+    ssize_t l;
 
     ret = gr = 0;
+    action = RETURN;
+
+    /* State switch */
+
     if (debug) 
 	fprintf(stderr, "eaptls: tls_state_machine, state %d\n", conv->state);
 
-    /* break on error, return on success */
-
     switch(conv->state) {
       case TLS_HANDSHAKE:
 	gr = gnutls_handshake(conv->session);
-	if (debug) 
-	    fprintf(stderr, "eaptls: gnutls_handshake returned %d: %s\n",
-		    gr, gnutls_strerror(gr));
-	if (gr >= 0) { 
-	    conv->state = gnutls_ia_handshake_p(conv->session) ? 
-			  TLS_IA : TLS_APP; 
-	    if (conv->state == TLS_APP) {
-		gr = gnutls_record_send(conv->session, "ABCD", 4);
-		if (debug) fprintf(stderr, "eaptls: gnutls_record_send returned %d: %s\n", gr, gnutls_strerror(gr));
-	    }
-	    return ret;
+	if (debug) fprintf(stderr, "eaptls: gnutls_handshake returned %d: %s\n", gr, gr <= 0 ? gnutls_strerror(gr) : "(n/a)");
+	if (gr >= 0) {
+	    action = gnutls_ia_handshake_p(conv->session) ? IA_RECV : APP_RECV;
+	    break;
 	}
-	if (gr == GNUTLS_E_INTERRUPTED || GNUTLS_E_AGAIN) return ret;
+	if (gr == GNUTLS_E_INTERRUPTED || GNUTLS_E_AGAIN) break;
+	action = ERROR;
 	break;
 
       case TLS_APP:
-	gr = gnutls_record_recv(conv->session, buf, sizeof(buf));
-	if (debug) 
-	    fprintf(stderr, "eaptls: gnutls_record_recv returned %d: %s\n", 
-		    gr, gnutls_strerror(gr));
-	if (gr < 0) break;
-	if (!gr) return ret;
-	if (debug) hexdumpfd(2, buf, gr, 0);
-	gr = gnutls_record_send(conv->session, buf, gr);  /* we simply echo */
-	if (debug) 
-	    fprintf(stderr, "eaptls: gnutls_record_send returned %d: %s\n", 
-		    gr, gnutls_strerror(gr));
-	return ret;
+	action = APP_RECV;
+	l = buf_maxget(&conv->totls_app);
+	if (!l) break;
+	gr = gnutls_record_send(conv->session, conv->totls_app.r, l); 
+	if (debug) fprintf(stderr, "eaptls: gnutls_record_send %d returned %d: %s\n", (int)l, gr, gr <= 0 ? gnutls_strerror(gr) : (gr < l ? "short write" : "all data sent"));
+	if (gr > 0) { if (gr > l) gr = l; buf_get(&conv->totls_app, gr); }
+	if (!gr || gr == GNUTLS_E_INTERRUPTED || GNUTLS_E_AGAIN) break;
+	action = ERROR;
+	break;
 
       default:
 	fprintf(stderr, "eaptls: Unknown state %d!\n", conv->state);
-	/* FIXME: ditch session */
-	return ret;
+	action = ERROR;
     }
 
-    /* Report GnuTLS error */
+    /* Action switch */
 
-    fprintf(stderr, "eaptls: "); gnutls_perror(gr);
+action:
+    if (debug) 
+	fprintf(stderr, "eaptls: tls_state_machine, action %d\n", action);
+
+    switch(action) {
+
+      /* Receive TLS data straight into reply packet buffer; keep header
+       * only if we received any data at all */
+
+      case APP_RECV:
+	conv->state = TLS_APP;
+	l = (repend - *rep) * sizeof(**rep) - 16;
+	if (l < 0) {
+	    fprintf(stderr, "eaptls: BUG: no room in OR reply buffer\n");
+	    action = ERROR; goto action;
+	}
+	o = *rep;
+	*o++ = net32(C_DS_INTERNAL);
+	*o++ = net32(META_VND_ANY);
+	*o++ = net32(C_DI_TLS_PAYLOAD);
+	*o++ = 0;					/* length done later */
+	gr = gnutls_record_recv(conv->session, o, l);
+	if (debug) {
+	    fprintf(stderr, "eaptls: gnutls_record_recv %d returned %d: %s\n", l, gr, gr < 0 ? gnutls_strerror(gr) : "data:");
+	    hexdumpfd(2, o, gr, 0);
+	}
+	if (gr > 0) {
+	    (*rep)[3] = net32(gr);	/* set length */
+	    gr = (gr + 3) >> 2;		/* length in 32-bit words */
+	    *rep = o + gr;
+	    ret += 16 + (gr << 2);	/* rounded length plus header */
+	    break;
+	}
+	if (gr == GNUTLS_E_INTERRUPTED || GNUTLS_E_AGAIN) break;
+	action = ERROR;
+	goto action;
+
+      case RETURN: 
+	break;
+
+      /* Error handling */
+
+      case PRERROR:
+	fprintf(stderr, "eaptls: "); gnutls_perror(gr);
+      case ERROR:
+      case DUMPSESS:
+	fprintf(stderr, "eaptls: FIXME: implement dump session action\n");
+	break;
+
+      default:
+	fprintf(stderr, "eaptls: Unknown action %d!\n", action);
+    }
     return ret;
 }
 
@@ -292,19 +353,20 @@ uint32_t tls_state_machine(uint32_t **rep, CONV_T *conv)
 
 int main(int argc, char **argv)
 {
+    static uint32_t rxbuf[(C_MAX_MSGSIZE >> 2) + 1];
+    static uint32_t txbuf[(C_MAX_MSGSIZE >> 2) + 1];
+    static CONV_T convring[MAX_CONVERSATIONS];
     static gnutls_certificate_credentials_t x509_cred;
     static gnutls_dh_params_t dh_params;
-    static uint32_t msgbuf[(C_MAX_MSGSIZE >> 2) + 1];
-    static CONV_T convring[MAX_CONVERSATIONS];
     CONV_T *conv;
+    STR_T s, tls_data, rx_payload;
     uint32_t spc, vnd, atr, len, *i, *e, *o; 
     uint32_t convr, convw, convnr;
-    uint32_t tls_flags, mtu, retint;
+    uint32_t rx_flags, tx_flags, mtu, rx_result, rx_action, retint;
     gnutls_datum dat;
-    STR_T s, tls_data;
     struct iovec iov[4];
-    ssize_t l;
     int parts, n;
+    ssize_t l;
 
     /* Initialise static data, parse options */
 
@@ -358,27 +420,28 @@ int main(int argc, char **argv)
          */
 
         /* Read header */
-        if (read(0, msgbuf, 8) != 8) { perror("eaptls: read"); break; }
-        if (net32(*msgbuf) != 0xbeefdead) {
-            fprintf(stderr, "eaptls: Invalid magic 0x%08x!\n", net32(*msgbuf)); 
+        if (read(0, rxbuf, 8) != 8) { perror("eaptls: read"); break; }
+        if (net32(*rxbuf) != 0xbeefdead) {
+            fprintf(stderr, "eaptls: Invalid magic 0x%08x!\n", net32(*rxbuf)); 
             break;
         }
-        len = net32(msgbuf[1]);
-        if (len < 8 || len > sizeof(msgbuf) - 4) {
+        len = net32(rxbuf[1]);
+        if (len < 8 || len > sizeof(rxbuf) - 4) {
             fprintf(stderr, "eaptls: Invalid length %d!\n", len); 
             break;
         }
 
         /* Read rest of message */
-        if (read(0, msgbuf + 2, len - 8) != len - 8) {
+        if (read(0, rxbuf + 2, len - 8) != len - 8) {
             perror("eaptls: read"); 
             break;
         }
 
 	/* Initialize defaults */
 
-	retint = convnr = tls_flags = ~0;
+	retint = convnr = rx_flags = tx_flags = rx_result = rx_action = ~0;
 	tls_data.l = 0;
+	rx_payload.l = 0;
 	mtu = 1230;
 
         /*
@@ -390,8 +453,8 @@ int main(int argc, char **argv)
          * fields of the EAP TLS packets.
          */
 
-        e = msgbuf + (len >> 2);
-        for(i = msgbuf + 2; i < e; i += ((len + 3) >> 2) + 4) {
+        e = rxbuf + (len >> 2);
+        for(i = rxbuf + 2; i < e; i += ((len + 3) >> 2) + 4) {
 
             /* Get space/vendor/attribute/length tuple */
 
@@ -402,25 +465,35 @@ int main(int argc, char **argv)
                         "%d, attribute %d, len %d\n", spc, vnd, atr, len);
             }
 
-            /* TLS-Data, -Flags attributes from EAP type-specific space */
+            /* Get TLS-Data, -Flags attributes from EAP type-specific space */
 
-            if (spc == rxspace && (vnd == 0 || vnd == META_ORD_ERR)) {
+            if (spc == rx_space && (vnd == 0 || vnd == META_ORD_ERR)) {
                 switch (atr) {
-                    case 0: tls_flags = net32(i[4]); break;
+                    case 0: rx_flags = net32(i[4]); break;
                     case 1: tls_data.p = (char *)&(i[4]); 
 			    tls_data.l = len;
                 }
                 continue;
             }
 
-            /* Transaction-Id (indexes timer ring containing conversations) */
+	    /* Get Transaction-Id (indexes timer ring containing
+	     * conversations); TLS-Payload (cleartext data to be transmitted
+	     * using as TLS Application Data) */
 
-            if (spc == C_DS_INTERNAL && atr == C_DI_TID) {
-                convnr = net32(i[4]);
-                continue;
-            }
+	    if (spc == C_DS_INTERNAL) switch(atr) {
+		case C_DI_TID:		convnr = net32(i[4]); 
+					continue;
+		case C_DI_TLS_PAYLOAD:	rx_payload.p = (char *)&(i[4]); 
+					rx_payload.l = len;
+					continue;
+		case C_DI_TLS_RESULT:	rx_result = net32(i[4]); 
+					continue;
+		case C_DI_TLS_ACTION:	rx_action = net32(i[4]); 
+					continue;
+	    }
 
-	    /* Framed-MTU, used as maximum amount of data to get from TLS */
+	    /* Get Framed-MTU, used as maximum amount of data to get from TLS at
+	     * a time */
 
             if (spc == C_DS_RAD_ATR && atr == C_DI_FRAMED_MTU) {
                 mtu = net32(i[4]);
@@ -430,18 +503,19 @@ int main(int argc, char **argv)
 
         /* Initialise reply buffer */
 
-        o = msgbuf; *o++ = net32(0xdeadbeef); o++;
-        e = msgbuf + sizeof(msgbuf) / sizeof(msgbuf[0]);
+        o = txbuf; *o++ = net32(0xdeadbeef); o++;
+        e = txbuf + sizeof(txbuf) / sizeof(txbuf[0]);
         len = 8; 
 
-	/* See if we actually got any TLS data */
+	/* See if we actually got any TLS attributes; flags must always be
+	 * present and cannot be 0xffffffff (at most 0xff) */
 
-	if (tls_flags == ~0) {
+	if (rx_flags == ~0) {
 	    errputcs("eaptls: no TLS attributes from core - wrong space?\n");
 	    retint = RET_NODATA; goto reply;
 	}
 
-        /* See if the request's conversation number is valid */
+        /* Obtain new conversation number if none given or given one invalid */
 
         if (convnr > MAX_CONVERSATIONS || !convring[convnr].inuse) {
 
@@ -461,12 +535,16 @@ int main(int argc, char **argv)
 	    conv = &convring[convnr];
 	    conv->inuse = 1;
 	    conv->state = TLS_HANDSHAKE;
-	    buf_init(&conv->fromtls, TLS_BUFFER_SIZE);
-	    buf_init(&conv->totls, TLS_BUFFER_SIZE);
+	    if (!buf_init(&conv->fromtls, TLS_BUFFER_SIZE) ||
+	        !buf_init(&conv->totls, TLS_BUFFER_SIZE) ||
+	        !buf_init(&conv->totls_app, TLS_BUFFER_SIZE)) {
+		errputcs("eaptls: no memory for new conversation!\n");
+		retint = RET_NOCONV; goto reply;
+	    }
 
-	    /* FIXME: add expiry time */
+	    /* FIXME: add expiry time to ring entry */
 
-	    /* Create Gnutls session - may be copied from existing one during
+	    /* Create Gnutls session - may be resumed from existing one during
 	     * handshake */
 
             gnutls_init(&conv->session, GNUTLS_SERVER);
@@ -482,38 +560,71 @@ int main(int argc, char **argv)
 	    fprintf(stderr, "eaptls: New conversation %u\n", convnr);
         }
 
-	/* Feed received data to TLS */
+	/* Skip length field in TLS data if length flag set */
+
+	if (rx_flags & C_DV_TLS_FLAGS_LENGTH) {
+	    tls_data.p += 4;
+	    tls_data.l -= 4;
+	}
+
+	/* Show what we got */
 
 	if (debug) {
-	    fprintf(stderr, "eaptls: Got %4d bytes of TLS data, flags 0x%02x, MTU %d, conversation %d\n", tls_data.l, tls_flags, mtu, convnr);
+	    fprintf(stderr, "eaptls: Got %4d bytes of TLS data, flags 0x%02x, MTU %d, conversation %d\n", tls_data.l, rx_flags, mtu, convnr);
 	    hexdumpfd(2, tls_data.p, tls_data.l, 0);
 	}
-	l = buf_maxput(&conv->totls);
-	if (tls_data.l > l) {
-	    fprintf(stderr, "eaptls: No room for %d bytes in buffer; has %d left!\n", tls_data.l, l);
 
-	    /* We should probably dump this session now */
-	    retint = RET_NOBUF; goto reply;
+	/* Process it */
+
+	if (rx_result != ~0) {
+	    conv->result = rx_result;	    /* Echoed when we drop the
+					       conversation */
 	}
-	memcpy(conv->totls.w, tls_data.p, tls_data.l);
-	buf_put(&conv->totls, tls_data.l);
+	if (tls_data.l) { 
 
-	/* Handle TLS according to state machine; may add attributes */
+	    /* Feed TLS data to TLS transport buffer, from where GnuTLS will
+	     * pull it using the gnutls_transport_read callback */
 
-	len += tls_state_machine(&o, conv);
+	    l = buf_maxput(&conv->totls);
+	    if (tls_data.l > l) {
+		fprintf(stderr, "eaptls: No room for %d bytes in transport "
+				"buffer; has %d left!\n", tls_data.l, l);
+
+		/* FIXME: We should dump this session now */
+		retint = RET_NOBUF; goto reply;
+	    }
+	    memcpy(conv->totls.w, tls_data.p, tls_data.l);
+	    buf_put(&conv->totls, tls_data.l);
+	}
+	if (rx_payload.l) {
+
+	    /* Feed payload data to TLS application buffer, where it's kept
+	     * until our state machine determines it's ready to send TLS
+	     * Application or InnerApplication records and it's all sent
+	     * in one go. */
+
+	    l = buf_maxput(&conv->totls_app);
+	    if (rx_payload.l > l) {
+		fprintf(stderr, "eaptls: No room for %d bytes in payload "
+				"buffer; has %d left!\n", rx_payload.l, l);
+
+		/* FIXME: We should dump this session now */
+		retint = RET_NOBUF; goto reply;
+	    }
+	    memcpy(conv->totls_app.w, rx_payload.p, rx_payload.l);
+	    buf_put(&conv->totls_app, rx_payload.l);
+	}
+
+	/* Work TLS engine according to our state machine, unless the 'more'
+	 * flag was set by the sender. May add attributes to response. */
+
+	if (!(rx_flags & C_DV_TLS_FLAGS_MORE))
+	    len += tls_state_machine(conv, &o, e, rx_action);
 
 	/* Add the given or new conversation number */
 
 	len += addint(&o, C_DI_TID, convnr);
 
-	/* Put flags from TLS in attribute 0 in txspace */
-
-	tls_flags = 0;
-	*o++ = net32(txspace); *o++ = net32(META_VND_ANY); *o++ = net32(0);
-	*o++ = net32(1);
-	*o++ = net32(tls_flags);
-	len += 20;
-
 	/* We're OK now */
 
 	retint = 0;
@@ -528,24 +639,58 @@ reply:
 
 	len += addint(&o, C_DI_INT, retint);
 
-	/* Add header for TLS data attribute, if we'll send any */
+	/* Check if we have any data to send. If the MTU is smaller than 256
+	 * or there's no data in the buffer or we have an error to report, we
+	 * won't send any data. */
 
 	l = 0;
-	if (retint < RET_ERROR && (l = buf_maxget(&conv->fromtls))) {
-	    if (l > mtu - 8) l = mtu - 8;
-	    *o++ = net32(txspace); *o++ = net32(META_VND_ANY); *o++ = net32(1);
-	    *o++ = net32(l);
+	if (!(rx_flags & C_DV_TLS_FLAGS_MORE) && mtu >= 256 &&
+	    retint < RET_ERROR && (l = buf_maxget(&conv->fromtls))) {
+
+	    /* If this is the first part of our transmission, set
+	     * length flag in tx_flags and add a length field at
+	     * the start of the TLS data. */
+
+	    tx_flags = 0;
+	    if (conv->fromtls.r == conv->fromtls.bu.p) {
+		tx_flags = C_DV_TLS_FLAGS_LENGTH;
+	    }
+
+	    /* If we can send less than our whole buffer, set the
+	     * more flag in tx_flags. */
+
+	    if (l > mtu - 12) {
+		l = mtu - 12;
+		tx_flags |= C_DV_TLS_FLAGS_MORE;
+	    }
+
+	    /* Add the flags */
+
+	    *o++ = net32(tx_space); *o++ = net32(META_VND_ANY); *o++ = net32(0);
+	    *o++ = net32(1);
+	    *o++ = net32(tx_flags);
+	    len += 20;
+
+	    /* Add a header for TLS transport length (optional) and data */
+
+	    *o++ = net32(tx_space); *o++ = net32(META_VND_ANY); *o++ = net32(1);
+	    if (tx_flags & C_DV_TLS_FLAGS_LENGTH) {
+		*o++ = net32(l + 4);
+		*o++ = net32(buf_maxget(&conv->fromtls));
+		len += 4;
+	    }
+	    else *o++ = net32(l);
 	    len += 16;
 	}
 
-	/* Prepare response */
+	/* Start to prepare response */
 
 	parts = 0;
-	iov[parts].iov_base = msgbuf;			    /* Header */
+	iov[parts].iov_base = txbuf;			    /* Header */
 	iov[parts].iov_len = len;
 	parts++;
 	if (l) {
-	    iov[parts].iov_base = conv->fromtls.r;	    /* TLS data */
+	    iov[parts].iov_base = conv->fromtls.r;	    /* TLS trans data */
 	    iov[parts].iov_len = l;
 	    parts++;
 	    len += l;
@@ -560,11 +705,12 @@ reply:
 
 	/* Send it */
 
-        msgbuf[1] = net32(len);
+        txbuf[1] = net32(len);
 	if (debug) {
-	    fprintf(stderr, "eaptls: replying, %d bytes in %d part(s):\n", len, parts);
+	    fprintf(stderr, "eaptls: replying, %d bytes in %d part(s):\n", 
+		    len, parts);
 	    for(n = 0; n < parts; n++)
-		hexdumpfd(2, iov[n].iov_base, iov[n].iov_len, n << 12);
+		hexdumpfd(2, iov[n].iov_base, iov[n].iov_len, 0);
 	}
 
         if (writev(1, iov, parts) != len) { perror("eaptls: write"); break; }
